import asyncio
from typing import Annotated

import typer
from rich.console import Console
from rich.text import Text

from issx.cli_utils import RichConfigReader
from issx.clients.gitlab import GitlabClient, GitlabInstanceClient
from issx.clients.redmine import RedmineClient, RedmineInstanceClient
from issx.domain import SupportedBackend
from issx.domain.config import InstanceConfig, ProjectFlatConfig
from issx.instance_managers.config_parser import GenericConfigParser
from issx.instance_managers.managers import InstanceManager
from issx.services import CopyIssueService

app = typer.Typer(no_args_is_help=True)
config_app = typer.Typer(name="config", no_args_is_help=True)
app.add_typer(config_app, name="config")

BackendOption = Annotated[SupportedBackend, typer.Option()]
InstanceNameOption = Annotated[str, typer.Option("--instance")]
ProjectOption = Annotated[
    str, typer.Option("--project", help="Project name configured in the config file")
]
console = Console()

InstanceManager.register_backend(
    SupportedBackend.gitlab, GitlabInstanceClient, GitlabClient
)
InstanceManager.register_backend(
    SupportedBackend.redmine, RedmineInstanceClient, RedmineClient
)


@app.command()
def copy(
    source_project_name: Annotated[
        str,
        typer.Option(
            "--source", help="Source project name configured in the config file"
        ),
    ],
    target_project_name: Annotated[
        str,
        typer.Option(
            "--target", help="Target project name configured in the config file"
        ),
    ],
    issue_id: int,
    title_format: Annotated[
        str,
        typer.Option(
            "--title-format",
            "-T",
            help="Template of a new issue title. Can contain placeholders of the"
            " issue attributes: {id}, {title}, {description}, {web_url}, {reference}",
        ),
    ] = "{title}",
    description_format: Annotated[
        str,
        typer.Option(
            "--description-format",
            "-D",
            help="Template of a new issue description. Can contain placeholders of the"
            " issue attributes: {id}, {title}, {description}, {web_url}, {reference}",
        ),
    ] = "{description}",
    allow_duplicates: Annotated[
        bool,
        typer.Option(
            "--allow-duplicates",
            "-A",
            help="Allow for duplicate issues. If set, the command will return the first"
            " issue found with the same title. If no issues are found,"
            " a new issue will be created.",
        ),
    ] = False,
    assign_to_me: Annotated[
        bool,
        typer.Option(
            "--assign-to-me",
            "-M",
            help="Whether to assign a newly created issue to the current user",
        ),
    ] = False,
) -> int:
    """Copy an issue from one project to another."""

    console.print(
        Text.assemble(
            f"Copying issue {issue_id} from project ",
            (source_project_name, "bold magenta"),
            " to project ",
            (target_project_name, "bold magenta"),
        )
    )
    config = GenericConfigParser.from_file()
    instance_manager = InstanceManager(config)
    try:
        source_client = instance_manager.get_project_client(source_project_name)
        target_client = instance_manager.get_project_client(target_project_name)
    except Exception as e:
        console.print_exception()
        console.print("Error when configuring client instance.\n", style="red")
        raise typer.Exit(1) from e
    new_issue = asyncio.run(
        CopyIssueService(source_client, target_client).copy(
            issue_id,
            title_format,
            description_format,
            allow_duplicates=allow_duplicates,
            assign_to_me=assign_to_me,
        )
    )
    console.print(f"Success!\n{new_issue}", style="green")
    return 0


@app.command()
def auth_verify(instance_name: InstanceNameOption) -> None:
    """Verify the authentication to the instance."""
    config = GenericConfigParser.from_file()
    instance_manager = InstanceManager(config)
    try:
        instance = instance_manager.get_instance_client(instance_name)
    except Exception as e:
        console.print_exception()
        console.print("Error when configuring client instance.\n", style="red")
        raise typer.Exit(1) from e
    try:
        username = asyncio.run(instance.auth())
        if not username:
            raise Exception("Authentication failed")
    except Exception as e:
        console.print_exception()
        console.print(
            f"Error when authenticating to {instance.get_instance_url()}",
            style="red",
        )
        raise typer.Exit(1) from e
    else:
        console.print(
            "Authentication successful",
            f"Instance: {instance.get_instance_url()}",
            f"User: {username}",
            sep="\n",
            style="green bold italic",
        )


@config_app.command()
def generate_instance(
    instance_name: Annotated[
        str,
        typer.Option(
            "--instance",
            help="Name of then instance to generate new_config for."
            " Only alphanumeric and -_ allowed",
        ),
    ],
) -> None:
    """Generate instance's new_config string"""

    new_config = RichConfigReader().read(InstanceConfig)
    console.print()
    console.print(Text(new_config.as_toml(f"projects.{instance_name}")))

    console.print(
        "\nSuccess!\nCopy the above config to your config file", style="green"
    )


@config_app.command()
def generate_project(
    project_name: Annotated[
        str,
        typer.Option(
            "--project",
            help="Name of then project to generate new_config for."
            " Only alphanumeric and -_ allowed",
        ),
    ],
) -> None:
    """Generate project's new_config string. Instance must be already configured."""

    new_config = RichConfigReader().read(ProjectFlatConfig)

    console.print()
    console.print(Text(new_config.as_toml(f"projects.{project_name}")))
    config = GenericConfigParser.from_file()

    try:
        config.get_instance_config(new_config.instance)
    except KeyError:
        console.print(
            f"\nWarning: instance [bold]{new_config.instance}[/bold] not found"
            f" in the config file. If you want to generate a missing config run"
            f" [bold]issx config generate-instance --instance {new_config.instance}"
            f"[/bold] command",
            style="yellow",
        )

    console.print(
        "\nSuccess!\nCopy the above config to your config file", style="green"
    )


if __name__ == "__main__":
    app()



class EntityDoesNotExistError(Exception):
    pass


class IssueDoesNotExistError(EntityDoesNotExistError):
    def __init__(self, issue_id: int):
        self.issue_id = issue_id
        super().__init__(f"Issue with id={issue_id} does not exist")


class ProjectDoesNotExistError(EntityDoesNotExistError):
    pass



from issx.clients.interfaces import IssueClientInterface
from issx.domain.issues import Issue


class CopyIssueService:
    def __init__(
        self, source_client: IssueClientInterface, target_client: IssueClientInterface
    ):
        self.source_client = source_client
        self.target_client = target_client

    async def copy(
        self,
        issue_id: int,
        title_format: str = "{title}",
        description_format: str = "{description}",
        allow_duplicates: bool = False,
        assign_to_me: bool = False,
    ) -> Issue:
        """
        Copy an issue from the source client to the target client optionally
        applying a title and description format. If `allow_duplicates` is `False`,
        the method will return a first issue found with the same title.

        :param issue_id: The ID of the issue to copy
        :param title_format: The format for the new issue title
        :param description_format: The format for the new issue description
        :param allow_duplicates: Whether to allow duplicate issues
        :param assign_to_me: Whether to assign the new issue to the current user
        :return: Newly created or existing issue in the target client
        """
        source_issue = await self.source_client.get_issue(issue_id)
        target_title = self._prepare_string(source_issue, title_format)
        if not allow_duplicates and (
            issues := await self.target_client.find_issues(target_title)
        ):
            return issues[0]
        new_issue = await self.target_client.create_issue(
            title=target_title,
            description=self._prepare_string(source_issue, description_format),
            assign_to_me=assign_to_me,
        )
        return new_issue

    @staticmethod
    def _prepare_string(issue: Issue, title_format: str) -> str:
        """
        :param issue: Issue object from which to create the new title
        :param title_format: template string for the new title.
        Can contain placeholders for the issue attributes:
        {id}, {title}, {description}, {web_url}, {reference}
        :return:
        """
        return title_format.format(
            id=issue.id,
            title=issue.title,
            description=issue.description,
            web_url=issue.web_url,
            reference=issue.reference,
        )



from issx.clients.interfaces import InstanceClientInterface, IssueClientInterface
from issx.domain import SupportedBackend
from issx.instance_managers.config_parser import GenericConfigParser


class InstanceManager:
    backends: dict[
        SupportedBackend,
        tuple[type[InstanceClientInterface], type[IssueClientInterface]],
    ] = {}

    def __init__(self, config: GenericConfigParser):
        self.config = config

    @classmethod
    def register_backend(
        cls,
        backend: SupportedBackend,
        instance_client_class: type[InstanceClientInterface],
        project_client_class: type[IssueClientInterface],
    ) -> None:
        cls.backends[backend] = (instance_client_class, project_client_class)

    @classmethod
    def clear_backends(cls) -> None:
        cls.backends.clear()

    def get_instance_client(self, instance: str) -> InstanceClientInterface:
        """
        Get an instance client for a given instance name.

        Converts the instance config to the appropriate config class
         and creates an instance client.
        Args:
            instance: Instance name

        Returns: Instance of an instance client
        """
        instance_config = self.config.get_instance_config(instance)
        client_class = self.backends[instance_config.backend][0]
        instance_config = client_class.instance_config_class(
            **instance_config.raw_config, raw_config=instance_config.raw_config
        )
        return client_class.instance_from_config(instance_config)

    def get_project_client(self, project: str) -> IssueClientInterface:
        """
        Get a project client for a given project name.

        Converts the project config to the appropriate config class
        and creates a project client.
        Args:
            project: Project name

        Returns: Instance of a project client

        """
        project_config = self.config.get_project_config(project)
        instance_config = self.config.get_instance_config(project_config.instance)
        instance_client_class, project_client_class = self.backends[
            instance_config.backend
        ]
        instance_config = instance_client_class.instance_config_class(
            **instance_config.raw_config, raw_config=instance_config.raw_config
        )
        project_config = project_client_class.project_config_class(
            **project_config.raw_config, raw_config=project_config.raw_config
        )
        return project_client_class.from_config(instance_config, project_config)



from enum import Enum
from typing import Any, TypeVar

from attr import Attribute
from rich.prompt import Prompt

from issx.domain.config import BaseConfig

TConfig = TypeVar("TConfig", bound=BaseConfig)


class RichConfigReader:
    def read(self, config_class: type[TConfig]) -> TConfig:
        fields = config_class.get_meaningful_fields()
        result: dict[Any, Any] = {}
        for field in fields:
            result[field.name] = self._read_field(field)
        return config_class(**result)

    @staticmethod
    def _read_field(field: Attribute) -> str:
        choices = (
            (
                [str(v) for v in field.type.__members__.values()]
                if issubclass(field.type, Enum)
                else None
            )
            if field.type
            else None
        )
        ask_kwargs = {}
        if field.default:
            ask_kwargs["default"] = str(field.default)
        return Prompt.ask(
            f"Enter [bold]{field.name}[/bold]",
            default=str(field.default) if field.default else ...,  # type: ignore[arg-type]
            choices=choices,
        )



import abc
from typing import ClassVar, Self

from issx.domain.config import InstanceConfig, ProjectFlatConfig
from issx.domain.issues import Issue


class InstanceClientInterface(abc.ABC):
    """
    Interface for a client that interacts with an issue tracker instance
    """

    instance_config_class: ClassVar[type[InstanceConfig]] = InstanceConfig

    @abc.abstractmethod
    async def auth(self) -> str | None:
        """
        Authenticate the client with the instance.
        Raises an internal error if the authentication fails or returns None.
        :return: The username of the authenticated user or None
        """
        pass

    @abc.abstractmethod
    def get_instance_url(self) -> str:
        """
        :return: The URL of the instance
        """
        pass

    @classmethod
    @abc.abstractmethod
    def instance_from_config(cls, instance_config: InstanceConfig) -> Self:
        """
        Create an instance of the client from a configuration dictionary.
        If required, the instance_config can be converted to an instance-specific
        configuration according to the `instance_config_class` attribute of the client.

        :param instance_config: The configuration object. Higher level code
        should validate the configuration.
        :return: An instance of the client
        """
        pass


class IssueClientInterface(abc.ABC):
    project_config_class: ClassVar[type[ProjectFlatConfig]] = ProjectFlatConfig

    @abc.abstractmethod
    async def get_issue(self, issue_id: int) -> Issue:
        """
        Retrieve an issue by its ID.
        Raises IssueDoesNotExistError if the issue does not exist.
        :param issue_id: The ID of the issue
        :return: Issue object
        """
        pass

    @abc.abstractmethod
    async def create_issue(
        self, title: str, description: str, assign_to_me: bool = False
    ) -> Issue:
        """
        Create a new issue.
        :param title: The title of the issue
        :param description: The description of the issue
        :param assign_to_me: Assign the issue to the authenticated user
        :return:
        """
        pass

    @abc.abstractmethod
    async def find_issues(self, title: str) -> list[Issue]:
        """
        Find issues by title using an exact match.
        :param title: The title of the issue
        :return: List of issues
        """
        pass

    @classmethod
    @abc.abstractmethod
    def from_config(
        cls, instance_config: InstanceConfig, project_config: ProjectFlatConfig
    ) -> Self:
        """
        Create an instance of the client from configuration classes.
        If required, the project_config can be converted to a project-specific
        configuration according to the `project_config_class` attribute of the client.

        Args:
            instance_config: InstanceConfig to configure the client to the instance
            project_config: ProjectFlatConfig to configure client. This config can then
            be converted to a project-specific config according to
            the `project_config_class` attribute of the client.
            to the particular project

        Returns:
            An instance of the client

        """
        pass



from collections.abc import Iterable

import attr
from attr import define

from issx.domain import SupportedBackend


@define(kw_only=True)
class BaseConfig:
    """
    Base class for config classes. It contains a raw_config attribute
    that should store the raw dictionary that was used to create the object.
    """

    raw_config: dict = attr.ib(
        validator=attr.validators.instance_of(dict),
        factory=dict,
    )

    @classmethod
    def get_meaningful_fields(cls) -> Iterable[attr.Attribute]:
        """
        Get all fields except the raw_config field.

        Returns: Iterable of fields
        """
        fields = attr.fields(cls)
        return [field for field in fields if field != fields.raw_config]

    def as_toml(self, key: str) -> str:
        """
        Convert the object to a TOML string.
        Args:
            key: Key to use for the TOML table

        Returns: TOML string
        """
        nl = "\n"
        header = f"[{key}]"
        attributes = [
            f"{field.name} = {repr(getattr(self, field.name))}"
            for field in self.get_meaningful_fields()
        ]
        return f"{header}\n{nl.join(attributes)}"


@define(kw_only=True)
class InstanceConfig(BaseConfig):
    backend: SupportedBackend = attr.ib(
        validator=attr.validators.in_(SupportedBackend.__members__.values())
    )
    url: str = attr.ib(validator=attr.validators.instance_of(str))
    token: str = attr.ib(validator=attr.validators.instance_of(str))


@define(kw_only=True)
class ProjectFlatConfig(BaseConfig):
    instance: str = attr.ib(validator=attr.validators.instance_of(str))
    project: str = attr.ib(validator=attr.validators.instance_of(str))



import tomllib
from pathlib import Path
from typing import Self

import attr
from attr import define, field

from issx.domain import SupportedBackend
from issx.domain.config import InstanceConfig, ProjectFlatConfig


@define
class GenericConfig:
    instances: dict[str, InstanceConfig]
    projects: dict[str, ProjectFlatConfig] = field()

    @projects.validator
    def validate_projects(
        self, attribute: attr.Attribute, value: dict[str, ProjectFlatConfig]
    ) -> None:
        for project in value.values():
            if project.instance not in self.instances:
                raise ValueError(f"Instance {project.instance} not found")


class GenericConfigParser:
    def __init__(self, data: GenericConfig):
        self._data: GenericConfig = data

    def get_instance_config(self, instance: str) -> InstanceConfig:
        return self._data.instances[instance]

    def get_project_config(self, project: str) -> ProjectFlatConfig:
        return self._data.projects[project]

    @classmethod
    def from_file(cls, config_file: Path | None = None) -> Self:
        config_file = config_file or cls._find_config_file()
        return cls.from_dict(tomllib.loads(config_file.read_text()))

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        parsed_data = GenericConfig(
            instances={
                name: InstanceConfig(
                    backend=SupportedBackend(instance.get("backend")),
                    url=instance.get("url"),
                    token=instance.get("token"),
                    raw_config=instance,
                )
                for name, instance in data["instances"].items()
            },
            projects={
                name: ProjectFlatConfig(
                    instance=project.get("instance"),
                    project=project.get("project"),
                    raw_config=project,
                )
                for name, project in data.get("projects", {}).items()
            },
        )
        return cls(parsed_data)

    @staticmethod
    def _find_config_file() -> Path:
        locations = [
            Path("issx.toml"),
            Path("~/.config/issx.toml").expanduser(),
        ]
        for location in locations:
            if location.exists() and location.is_file() and location.suffix == ".toml":
                return location
        raise FileNotFoundError("No configuration file found")
